"""
The Go starter template.


Author: Tom Fleet
Created: 29/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path

import aiofiles
import aiofiles.os

from pytoil.exceptions import GoNotInstalledError

GO = shutil.which("go")


class GoStarter:
    def __init__(self, path: Path, name: str, go: str | None = GO) -> None:
        self.path = path
        self.name = name
        self.go = go
        self.root = self.path.joinpath(self.name).resolve()
        self.files = [
            self.root.joinpath(filename) for filename in ["README.md", "main.go"]
        ]

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(path={self.path!r}, name={self.name!r}, go={self.go!r})"
        )

    __slots__ = ("path", "name", "go", "root", "files")

    async def generate(self, username: str | None = None) -> None:
        """
        Generate a new Go starter template.
        """
        if not self.go:
            raise GoNotInstalledError

        await aiofiles.os.mkdir(self.root)

        # Call go mod init
        proc = await asyncio.create_subprocess_exec(
            self.go,
            "mod",
            "init",
            f"github.com/{username}/{self.name}",
            cwd=self.root,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        await proc.wait()

        for file in self.files:
            file.touch()

        # Put the header in the README
        readme = self.root.joinpath("README.md")
        async with aiofiles.open(readme, mode="w", encoding="utf-8") as f:
            await f.write(f"# {self.name}\n")

        # Populate the go file
        main_go = self.root.joinpath("main.go")
        async with aiofiles.open(main_go, mode="w", encoding="utf-8") as f:
            await f.write(
                'package main\n\nimport "fmt"\n\nfunc main() {\n\tfmt.Println("Hello'
                ' World")\n}\n'
            )



"""
The pytoil remove command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncio
import functools
import shutil

import asyncclick as click
import questionary
from wasabi import msg

from pytoil.config import Config


@click.command()
@click.argument("projects", nargs=-1)
@click.option("-f", "--force", is_flag=True, help="Force delete without confirmation.")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all of your local projects.",
)
@click.pass_obj
async def remove(
    config: Config, projects: tuple[str, ...], force: bool, all_: bool
) -> None:
    """
    Remove projects from your local filesystem.

    The remove command provides an easy interface for decluttering your local
    projects directory.

    You can selectively remove any number of projects by passing them as
    arguments or nuke the whole lot with "--all/-a" if you want.

    As with most programmatic deleting, the directories are deleted instantly and
    not sent to trash. As such, pytoil will prompt you for confirmation before
    doing anything.

    The "--force/-f" flag can be used to force deletion without the confirmation
    prompt. Use with caution!

    Examples:

    $ pytoil remove project1 project2 project3

    $ pytoil remove project1 project2 project3 --force

    $ pytoil remove --all

    $ pytoil remove --all --force
    """
    local_projects: set[str] = {
        f.name
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    if not local_projects:
        msg.warn("You don't have any local projects to remove", exits=1)

    if not projects and not all_:
        msg.warn(
            "If not using the '--all' flag, you must specify projects to remove.",
            exits=1,
        )

    # If user gives a project that doesn't exist (e.g. typo), abort
    for project in projects:
        if project not in local_projects:
            msg.warn(
                f"{project!r} not found under {config.projects_dir}. Was it a typo?",
                exits=1,
            )

    to_delete = local_projects if all_ else projects

    if not force:
        if all_:
            question = questionary.confirm(
                "This will delete ALL of your projects. Are you sure?",
                default=False,
                auto_enter=False,
            )
        elif len(projects) <= 3:
            # Nice number to show the names
            question = questionary.confirm(
                f"This will delete {', '.join(projects)} from your local filesystem."
                " Are you sure?",
                default=False,
                auto_enter=False,
            )
        else:
            # Too many to print the names nicely
            question = questionary.confirm(
                f"This will delete {len(projects)} projects from your local filesystem."
                " Are you sure?",
                default=False,
                auto_enter=False,
            )

        confirmed: bool = await question.ask_async()

        if not confirmed:
            raise click.Abort()

    # If we get here, user has used --force or said yes when prompted
    # Go async!
    await asyncio.gather(*[remove_and_report(config, project) for project in to_delete])


async def remove_and_report(config: Config, project: str) -> None:
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(
        executor=None,
        func=functools.partial(
            shutil.rmtree,
            path=config.projects_dir.joinpath(project),
            ignore_errors=True,
        ),
    )
    msg.good(f"Deleted {project}")



"""
The pytoil info command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncclick as click
from rich.console import Console
from rich.table import Table, box
from wasabi import msg

from pytoil.api import API
from pytoil.config import Config
from pytoil.exceptions import RepoNotFoundError
from pytoil.repo import Repo


@click.command()
@click.argument("project", nargs=1)
@click.pass_obj
async def info(config: Config, project: str) -> None:
    """
    Get useful info for a project.

    Given a project name (can be local or remote), 'info' will return a summary
    description of the project.

    If the project is on GitHub, info will prefer getting information from the GitHub
    API as this is more detailed.

    If the project is local only, some information is extracted from the operating
    system about the project.

    Examples:

    $ pytoil info my_project
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        info = await repo.info(api)
    except RepoNotFoundError:
        msg.warn(f"{project!r} not found locally or on GitHub. Was it a typo?", exits=1)
    else:
        click.secho(f"Info for {project}:", fg="cyan", bold=True)

        table = Table(box=box.SIMPLE)
        table.add_column("Key", style="cyan", justify="right")
        table.add_column("Value", justify="left")

        for key, val in info.items():
            table.add_row(f"{key}:", str(val))

        console = Console()
        console.print(table)



"""
The rust starter template.


Author: Tom Fleet
Created: 29/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path

import aiofiles
import aiofiles.os

from pytoil.exceptions import CargoNotInstalledError

CARGO = shutil.which("cargo")


class RustStarter:
    def __init__(self, path: Path, name: str, cargo: str | None = CARGO) -> None:
        self.path = path
        self.name = name
        self.cargo = cargo
        self.root = self.path.joinpath(self.name).resolve()
        self.files = [self.root.joinpath(filename) for filename in ["README.md"]]

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(path={self.path!r}, name={self.name!r}, cargo={self.cargo!r})"
        )

    __slots__ = ("path", "name", "cargo", "root", "files")

    async def generate(self, username: str | None = None) -> None:
        """
        Generate a new rust/cargo starter template.
        """
        if not self.cargo:
            raise CargoNotInstalledError

        await aiofiles.os.mkdir(self.root)

        # Call cargo init
        proc = await asyncio.create_subprocess_exec(
            self.cargo,
            "init",
            "--vcs",
            "none",
            cwd=self.root,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        await proc.wait()

        # Create the README
        for file in self.files:
            file.touch()

        readme = self.root.joinpath("README.md")
        async with aiofiles.open(readme, mode="w", encoding="utf-8") as f:
            await f.write(f"# {self.name}\n")



"""
Module responsible for handling python virtual environments
through the std lib `venv` module.


Author: Tom Fleet
Created: 24/12/2021
"""

from __future__ import annotations

import asyncio
import functools
import sys
from pathlib import Path
from typing import Sequence

import aiofiles.os
import virtualenv


class Venv:
    root: Path

    def __init__(self, root: Path) -> None:
        self.root = root

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r})"

    __slots__ = ("root",)

    @property
    def project_path(self) -> Path:
        return self.root.resolve()

    @property
    def executable(self) -> Path:
        return self.project_path.joinpath(".venv/bin/python")

    @property
    def name(self) -> str:
        return "venv"

    async def exists(self) -> bool:
        """
        Checks whether the virtual environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then both the project and virtual environment
        must also exist and therefore must be valid.
        """
        return await aiofiles.os.path.exists(self.executable)

    async def create(
        self, packages: Sequence[str] | None = None, silent: bool = False
    ) -> None:
        """
        Create the virtual environment in the project.

        If packages are specified here, these will be installed
        once the environment is created.

        Args:
            packages (Optional[List[str]], optional): Packages to install immediately
                after environment creation. Defaults to None.
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        # TODO: I'm actually not sure how to mock this out?
        # maybe we just create a tempdir and create a real venv, check the contents
        # then delete it? Would slow down the tests though
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(
            executor=None,
            func=functools.partial(
                virtualenv.cli_run,
                args=[str(self.project_path.joinpath(".venv")), "--quiet"],
            ),
        )

        # Install any specified packages
        if packages:
            await self.install(packages=packages, silent=silent)

    async def install(self, packages: Sequence[str], silent: bool = False) -> None:
        """
        Generic `pip install` method.

        Takes a list of packages to install. All packages are passed through to pip
        so any versioning syntax will work as expected.

        Args:
            packages (List[str]): List of packages to install, if only 1 package
                still must be a list e.g. `["black"]`.
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        proc = await asyncio.create_subprocess_exec(
            f"{self.executable}",
            "-m",
            "pip",
            "install",
            *packages,
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def install_self(self, silent: bool = False) -> None:
        """
        Installs current package.

        We first try the equivalent of `pip install -e .[dev]` as a large
        number of packages declare a [dev] extra which contains everything
        needed to work on it.

        Pip will automatically fall back to `pip install -e .` in the event
        `.[dev]` does not exist and every python package must know how to
        install itself this way by definition.

        Args:
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        # Before installing the package, ensure a virtualenv exists
        if not await self.exists():
            await self.create(silent=silent)

        # We try .[dev] first as most packages I've seen have this
        # and pip will automatically fall back to '.' if not
        proc = await asyncio.create_subprocess_exec(
            f"{self.executable}",
            "-m",
            "pip",
            "install",
            "-e",
            ".[dev]",
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()



__version__ = "0.23.2"

__doc__ = """

Helpful CLI to automate the development workflow.

- Create and manage your local and remote projects

- Build projects from cookiecutter templates.

- Easily create/manage virtual environments.

- Minimal configuration required.
"""



"""
Interface that all virtual environment classes must
satisfy.


Author: Tom Fleet
Created: 24/12/2021
"""

from __future__ import annotations

from pathlib import Path
from typing import Protocol, Sequence


class Environment(Protocol):
    @property
    def project_path(self) -> Path:
        """
        `.project_path` represents the root directory of the
        project associated with the virtual environment.
        """
        ...

    @property
    def executable(self) -> Path:
        """
        `.executable` is the absolute path to the virtual environment's
        python interpreter.
        """
        ...

    @property
    def name(self) -> str:
        """
        Returns the type of environment implemented by the concrete instance.
        Used for logging and debugging

        Returns:
            str: E.g. 'conda', 'venv', 'poetry' etc.
        """
        ...

    async def exists(self) -> bool:
        """
        `.exists()` checks whether the virtual environment exists.
        How it does this is up to the concrete implementation, but a good
        example might be simply checking if `self.executable` exists.
        """
        ...

    async def create(
        self, packages: Sequence[str] | None = None, silent: bool = False
    ) -> None:
        """
        Method to create the virtual environment. If packages are specified,
        these can be installed during environment creation.
        """
        ...

    async def install(self, packages: Sequence[str], silent: bool = False) -> None:
        """
        Generic install method.

        Installs `packages` into the correct virtual environment.

        Args:
            packages (List[str]): List of valid packages to install.
            silent (bool, optional): Whether to discard or display output.
        """
        ...

    async def install_self(self, silent: bool = False) -> None:
        """
        Installs the current project.

        For example: `pip install -e .[dev]` or `poetry install`
        """
        ...



"""
Global defaults for pytoil.

Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import os
from pathlib import Path

# Default path for pytoil's config file
CONFIG_FILE: Path = Path.home().joinpath(".pytoil.toml").resolve()

# Valid pytoil config keys
CONFIG_KEYS: set[str] = {
    "projects_dir",
    "token",
    "username",
    "vscode",
    "common_packages",
    "git",
}

# API cache
CACHE_DIR = Path.home().joinpath(".cache/pytoil")
CACHE_TIMEOUT_SECS = 120


# Pytoil docs URL
PYTOIL_DOCS_URL: str = "https://followtheprocess.github.io/pytoil/"

# Defaults for pytoil config
PROJECTS_DIR: Path = Path.home().joinpath("Development").resolve()
TOKEN: str = os.getenv("GITHUB_TOKEN", "")
USERNAME: str = ""
VSCODE: bool = False
CODE_BIN: str = "code"
COMMON_PACKAGES: list[str] = []
GIT: bool = True

# Config Schema
CONFIG_SCHEMA = """

# The .pytoil.toml config file

## projects_dir *(str)*

The absolute path to where you keep your development projects
(e.g. /Users/you/Projects).

## token *(str)*

Your GitHub personal access token. This must have a minimum of repo read access. See the documentation here:
https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/creating-a-personal-access-token

Pytoil will try and get this from the config file initially, then fall back to the $GITHUB_TOKEN environment
variable. If neither of these places are set, you will not be able to use pytoil commands that rely on the
GitHub API. Pytoil will notify you of this when any of these commands are called.

## username *(str)*

Your GitHub username. Pytoil needs this so it can construct urls to your projects.

## vscode *(bool)*

Whether you want pytoil to open projects up using VSCode. This will happen on 'new' and 'checkout'.

## code_bin *(str)*

The name of the VSCode binary. Will differ depending on version installed. Either "code" (default)
or "code-insiders"

## common_packages *(List[str])*

A list of python packages to inject into every virtual environment pytoil creates
(e.g. linters, formatters and other dev dependencies).

Any versioning syntax (e.g. mypy>=0.902) will work as expected here as these packages
are passed straight through to installation tools like pip and conda.

## git *(bool)*

Whether or not you want pytoil to create an empty git repo when you make a new project with
'pytoil new'. This can also be disabled on a per use basis using the '--no-git' flag.
"""



"""
Module responsible for handling pytoil's interface with the
GitHub GraphQL v4 API.


Author: Tom Fleet
Created: 21/12/2021
"""


from __future__ import annotations

from datetime import datetime
from typing import Any

import aiofiles.os
import httpx
import httpx_cache
import humanize

from pytoil import __version__
from pytoil.api import queries
from pytoil.config import defaults

URL = "https://api.github.com/graphql"
GITHUB_TIME_FORMAT = r"%Y-%m-%dT%H:%M:%SZ"
DEFAULT_REPO_LIMIT = 50


class API:
    def __init__(self, username: str, token: str, url: str = URL) -> None:
        """
        Container for methods and data for hitting the GitHub v4
        GraphQL API

        Args:
            username (str): User's GitHub username.
            token (str): User's personal access token.
            url (str, optional): GraphQL URL
                defaults to https://api.github.com/graphql
        """
        self.username = username
        self.token = token
        self.url = url

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(username={self.username}, token={self.token}, url={self.url})"
        )

    __slots__ = ("username", "token", "url")

    @property
    def headers(self) -> dict[str, str]:
        return {
            "Authorization": f"token {self.token}",
            "User-Agent": f"pytoil/{__version__}",
            "Accept": "application/vnd.github.v4+json",
            "Cache-Control": f"max-age:{defaults.CACHE_TIMEOUT_SECS}",
        }

    async def get_repos(
        self, limit: int = DEFAULT_REPO_LIMIT
    ) -> list[dict[str, Any]] | None:
        """
        Gets some summary info for all the users repos.

        Args:
            limit (int, optional): Maximum number of repos to return.
                Defaults to DEFAULT_REPO_LIMIT.

        Returns:
            list[dict[str, Any]]: The repos info.
        """
        # TODO: Not sure I like cache path stuff being here?
        cache_dir = defaults.CACHE_DIR.joinpath("get_repos")
        if not await aiofiles.os.path.exists(cache_dir):
            await aiofiles.os.makedirs(cache_dir)

        async with httpx.AsyncClient(
            http2=True,
            headers=self.headers,
            transport=httpx_cache.AsyncCacheControlTransport(
                cacheable_methods=("POST",),
                cache=httpx_cache.FileCache(
                    cache_dir=defaults.CACHE_DIR.joinpath("get_repos")
                ),
            ),
        ) as client:
            r = await client.post(
                self.url,
                json={
                    "query": queries.GET_REPOS,
                    "variables": {"username": self.username, "limit": limit},
                },
            )
            r.raise_for_status()

        raw: dict[str, Any] = r.json()

        if data := raw.get("data"):
            return list(data["user"]["repositories"]["nodes"])

        return None

    async def get_repo_names(self, limit: int = DEFAULT_REPO_LIMIT) -> set[str]:
        """
        Gets the names of all repos owned by the authenticated user.

        Args:
            limit (int, optional): Maximum number of repos to return.
                Defaults to DEFAULT_REPO_LIMIT.

        Raises:
            ValueError: If the GraphQL query is malformed.

        Returns:
            Set[str]: The names of the user's repos.
        """
        # TODO: Not sure I like cache path stuff being here?
        cache_dir = defaults.CACHE_DIR.joinpath("get_repo_names")
        if not await aiofiles.os.path.exists(cache_dir):
            await aiofiles.os.makedirs(cache_dir)

        async with httpx.AsyncClient(
            http2=True,
            headers=self.headers,
            transport=httpx_cache.AsyncCacheControlTransport(
                cacheable_methods=("POST",),
                cache=httpx_cache.FileCache(cache_dir=cache_dir),
            ),
        ) as client:
            r = await client.post(
                self.url,
                json={
                    "query": queries.GET_REPO_NAMES,
                    "variables": {"username": self.username, "limit": limit},
                },
            )
            r.raise_for_status()

        raw: dict[str, Any] = r.json()

        # TODO: I don't like the indexing here, must be a more type safe way of doing this
        # What happens when there are no nodes? e.g. user has no forks
        if data := raw.get("data"):
            return {node["name"] for node in data["user"]["repositories"]["nodes"]}

        else:
            raise ValueError(f"Bad GraphQL: {raw}")

    async def get_forks(
        self, limit: int = DEFAULT_REPO_LIMIT
    ) -> list[dict[str, Any]] | None:
        """
        Gets info for all users forks.

        Args:
            limit: (int, optional): Maximum number of repos to return.
                Defaults to DEFAULT_REPO_LIMIT.

        Returns:
            list[dict[str, Any]]: The JSON info for all forks.
        """
        # TODO: Not sure I like cache path stuff being here?
        cache_dir = defaults.CACHE_DIR.joinpath("get_forks")
        if not await aiofiles.os.path.exists(cache_dir):
            await aiofiles.os.makedirs(cache_dir)

        async with httpx.AsyncClient(
            http2=True,
            headers=self.headers,
            transport=httpx_cache.AsyncCacheControlTransport(
                cacheable_methods=("POST",),
                cache=httpx_cache.FileCache(cache_dir=cache_dir),
            ),
        ) as client:
            r = await client.post(
                self.url,
                json={
                    "query": queries.GET_FORKS,
                    "variables": {"username": self.username, "limit": limit},
                },
            )
            r.raise_for_status()

        raw: dict[str, Any] = r.json()

        if data := raw.get("data"):
            return list(data["user"]["repositories"]["nodes"])

        return None

    async def check_repo_exists(self, name: str) -> bool:
        """
        Checks whether or not a repo given by `name` exists
        under the current user

        Args:
            name (str): Repo name to check for

        Returns:
            bool: True if repo exists on GitHub, else False.
        """
        # Note: we don't cache this response as we want the data to be up to date always
        async with httpx.AsyncClient(http2=True, headers=self.headers) as client:
            r = await client.post(
                self.url,
                json={
                    "query": queries.CHECK_REPO_EXISTS,
                    "variables": {"username": self.username, "name": name},
                },
            )
            r.raise_for_status()

        raw: dict[str, Any] = r.json()

        if data := raw.get("data"):
            if data["repository"] is None:
                return False
            else:
                return True
        else:
            raise ValueError(f"Bad GraphQL: {raw}")

    async def create_fork(self, owner: str, repo: str) -> None:
        """
        Use the v3 REST API to create a fork of the specified repository
        under the authenticated user.

        Args:
            owner (str): Owner of the original repo.
            repo (str): Name of the original repo.
        """
        rest_headers = self.headers.copy()
        rest_headers["Accept"] = "application/vnd.github.v3+json"
        fork_url = f"https://api.github.com/repos/{owner}/{repo}/forks"

        # Nothing to cache here
        async with httpx.AsyncClient(http2=True, headers=rest_headers) as client:
            r = await client.post(fork_url)
            r.raise_for_status()

    @staticmethod
    def _humanize_datetime(dt: str) -> str:
        """
        Takes a string datetime of GITHUB_TIME_FORMAT
        and converts it to our STR_TIME_FORMAT.
        """
        s: str = humanize.naturaltime(datetime.strptime(dt, GITHUB_TIME_FORMAT))
        return s

    async def get_repo_info(self, name: str) -> dict[str, Any] | None:
        """
        Gets some descriptive info for the repo given by
        `name` under the current user.

        Args:
            name (str): Name of the repo to fetch info for.

        Returns:
            Dict[str, Any]: Repository info.
        """
        # TODO: Not sure I like cache path stuff being here?
        # Cache each name separately to avoid collision
        cache_dir = defaults.CACHE_DIR.joinpath(f"get_repo_info/{name}")
        if not await aiofiles.os.path.exists(cache_dir):
            await aiofiles.os.makedirs(cache_dir)

        async with httpx.AsyncClient(
            http2=True,
            headers=self.headers,
            transport=httpx_cache.AsyncCacheControlTransport(
                cacheable_methods=("POST",),
                cache=httpx_cache.FileCache(cache_dir=cache_dir),
            ),
        ) as client:
            r = await client.post(
                self.url,
                json={
                    "query": queries.GET_REPO_INFO,
                    "variables": {"username": self.username, "name": name},
                },
            )
            r.raise_for_status()

        raw: dict[str, Any] = r.json()

        if data := raw.get("data"):
            if repo := data.get("repository"):
                return {
                    "Name": repo["name"],
                    "Description": repo["description"],
                    "Created": self._humanize_datetime(repo["createdAt"]),
                    "Updated": self._humanize_datetime(repo["updatedAt"]),
                    "Size": humanize.naturalsize(
                        int(repo["diskUsage"]) * 1024
                    ),  # diskUsage is in kB
                    "License": repo["licenseInfo"]["name"]
                    if repo.get("licenseInfo")
                    else None,
                    "Remote": True,
                }
            return None
        return None



"""
The async-pytoil config command group.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncclick as click
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table, box
from wasabi import msg

from pytoil.config import Config, defaults


@click.group()
async def config() -> None:
    """
    Interact with pytoil's configuration.

    The config command group allows you to get, show and explain pytoil's configuration.
    """


@config.command()
@click.pass_obj
async def show(config: Config) -> None:
    """
    Show pytoil's config.

    The show command allows you to easily see pytoil's current config.

    The values are taken directly from the config file where specified or
    the defaults otherwise.

    Examples:

    $ pytoil config show
    """
    table = Table(box=box.SIMPLE)
    table.add_column("Key", style="cyan", justify="right")
    table.add_column("Value", justify="left")

    for key, val in config.to_dict().items():
        table.add_row(f"{key}:", str(val))

    console = Console()
    console.print(table)


@config.command()
@click.argument("key", nargs=1)
@click.pass_obj
async def get(config: Config, key: str) -> None:
    """
    Get the currently set value for a config key.

    The get command will only allow valid pytoil config keys.

    Examples:

    $ pytoil config get vscode
    """
    if key not in defaults.CONFIG_KEYS:
        msg.warn(f"{key!r} is not a valid pytoil config key.", exits=1)

    config_dict = config.to_dict()
    # Make the key a nice colour
    config_start = click.style(f"{key}: ", fg="cyan")
    config_msg = config_start + f"{config_dict.get(key)!r}"
    click.secho(config_msg)


@config.command()
@click.pass_obj
async def edit(config: Config) -> None:
    """
    Open pytoil's config file in $EDITOR.

    Examples:

    $ pytoil config edit
    """
    click.edit(filename=defaults.CONFIG_FILE)


@config.command()
async def explain() -> None:
    """
    Print a list and description of pytoil config values.

    Examples:

    $ pytoil config explain
    """
    console = Console()
    markdown = Markdown(defaults.CONFIG_SCHEMA, justify="center")
    console.print(markdown)



"""
Module responsible for pytoil's interface with VSCode.


Author: Tom Fleet
Created: 23/12/2021
"""

from __future__ import annotations

import asyncio
import json
import shutil
import sys
from pathlib import Path

import aiofiles
import aiofiles.os

from pytoil.exceptions import CodeNotInstalledError

# From VSCode 1.57.1 the 'python.pythonPath' setting is being
# deprecated in favour of 'python.defaultInterpreterPath'
# it works in a slightly different way but for us it's effectively
# a straight swap
WORKSPACE_PYTHON_SETTING = "python.defaultInterpreterPath"

CODE = shutil.which("code")


class VSCode:
    def __init__(self, root: Path, code: str | None = CODE) -> None:
        self.root = root
        self.code = code

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r}, code={self.code!r})"

    __slots__ = ("root", "code")

    @property
    def workspace_settings(self) -> Path:
        """
        Return the absolute path to the project's root
        VSCode workspace settings file.
        """
        return self.root.joinpath(".vscode/settings.json")

    async def open(self) -> None:
        """
        Opens a project root in VSCode using the configured
        code binary command.
        """
        if not self.code:
            raise CodeNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.code,
            ".",
            cwd=self.root,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        await proc.wait()

    async def set_workspace_python(self, python_path: Path) -> None:
        """
        Sets the VSCode workspace setting `python.defaultInterpreterPath`
        to `python_path`, overwriting if already exists.

        Args:
            python_path (Path): The path to the virtual environment's
                python interpreter.
        """
        # Take care not to resolve `python_path` as it will resolve the symlink
        # back to the system python. As such, `python_path` should be the executable
        # from a `VirtualEnv` or a `CondaEnv` class to ensure this is safely done

        new_settings_dict: dict[str, str] = {WORKSPACE_PYTHON_SETTING: str(python_path)}

        # 2 cases to worry about here:
        # 1) File doesn't exist -> easy, just write new ones
        # 2) File exists -> read in what is already there, add new settings, write back overwriting

        if not await aiofiles.os.path.exists(self.workspace_settings):
            # Create and write new ones
            # The entire .vscode folder might not exist
            await aiofiles.os.makedirs(self.workspace_settings.parent, exist_ok=True)
            async with aiofiles.open(
                self.workspace_settings, mode="w", encoding="utf-8"
            ) as file:
                content = json.dumps(new_settings_dict)
                await file.write(content)

        else:
            # File exists, preserve any content
            async with aiofiles.open(
                self.workspace_settings, mode="r", encoding="utf-8"
            ) as file:
                content = await file.read()
                settings: dict[str, str] = json.loads(content)

            settings.update(new_settings_dict)

            # Write the new settings back
            async with aiofiles.open(
                self.workspace_settings, mode="w", encoding="utf-8"
            ) as file:
                content = json.dumps(settings)
                await file.write(content)



"""
Module responsible for interacting with git via subprocesses.


Author: Tom Fleet
Created: 22/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path

from pytoil.exceptions import GitNotInstalledError

GIT = shutil.which("git")


class Git:
    def __init__(self, git: str | None = GIT) -> None:
        self.git = git

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(git={self.git!r})"

    __slots__ = ("git",)

    async def clone(self, url: str, cwd: Path, silent: bool = True) -> None:
        """
        Clone a repo.

        Args:
            url (str): The clone url of the repo
            cwd (Path): The cwd under which to clone
            silent (bool, optional): Whether to hook the output
                up to stdout and stderr (False) or to discard and keep silent (True).
                Defaults to True.
        """
        if not self.git:
            raise GitNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.git,
            "clone",
            url,
            cwd=cwd,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def init(self, cwd: Path, silent: bool = True) -> None:
        """
        Initialise a new git repo.

        Args:
            cwd (Path): The cwd to initialise the repo in.
            silent (bool, optional): Whether to hook the output
                up to stdout and stderr (False) or to discard and keep silent (True).
                Defaults to True.
        """
        if not self.git:
            raise GitNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.git,
            "init",
            cwd=cwd,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def add(self, cwd: Path, silent: bool = True) -> None:
        """
        Stages all files in cwd.

        Args:
            cwd (Path): The cwd to stage all child files in.
            silent (bool, optional): Whether to hook the output up
                to stdout and stderr (False) or to discard and keep silent (True).
                Defaults to True.
        """
        if not self.git:
            raise GitNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.git,
            "add",
            "-A",
            cwd=cwd,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def commit(
        self,
        cwd: Path,
        message: str = "Initial Commit (Automated at Project Creation)",
        silent: bool = True,
    ) -> None:
        """
        Commits the current state.

        Args:
            cwd (Path): cwd of the repo to commit.
            message (str, optional): Optional commit message.
                Defaults to "Initial Commit (Automated at Project Creation)"
            silent (bool, optional): Whether to hook the output up
                to stdout and stderr (False) or to discard and keep silent (True).
                Defaults to True.
        """
        if not self.git:
            raise GitNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.git,
            "commit",
            "-m",
            message,
            cwd=cwd,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def set_upstream(
        self, owner: str, repo: str, cwd: Path, silent: bool = True
    ) -> None:
        """
        Sets the upstream repo for a local repo, e.g. on a cloned fork.

        Note difference between origin and upstream, origin of a cloned fork
        would be user/forked_repo where as upstream would be original_user/forked_repo.

        Args:
            owner (str): Owner of the upstream repo.
            repo (str): Name of the upstream repo (typically the same as the fork)
            cwd (Path): Root of the project where the local git repo is.
            silent (bool, optional): Whether to hook the output
                up to stdout and stderr (False) or to discard and keep silent (True).
                Defaults to True.
        """
        if not self.git:
            raise GitNotInstalledError

        base_url = "https://github.com"
        constructed_upstream = f"{base_url}/{owner}/{repo}.git"

        proc = await asyncio.create_subprocess_exec(
            self.git,
            "remote",
            "add",
            "upstream",
            constructed_upstream,
            cwd=cwd,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()



"""
Module responsible for handling python environments
with a `requirements.txt` (or `requirements_dev.txt`)


Author: Tom Fleet
Created: 26/12/2021
"""

from __future__ import annotations

import asyncio
import sys
from pathlib import Path

import aiofiles.os

from pytoil.environments.virtualenv import Venv


class Requirements(Venv):
    def __init__(self, root: Path) -> None:
        self.root = root
        super().__init__(root)

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r})"

    __slots__ = ("root",)

    @property
    def name(self) -> str:
        return "requirements file"

    async def install_self(self, silent: bool = False) -> None:
        """
        Installs everything in the requirements file into
        a python environment.
        """
        if not await self.exists():
            await self.create(silent=silent)

        requirements_file = "requirements.txt"

        if await aiofiles.os.path.exists(
            self.project_path.joinpath("requirements_dev.txt")
        ):
            requirements_file = "requirements_dev.txt"

        proc = await asyncio.create_subprocess_exec(
            f"{self.executable}",
            "-m",
            "pip",
            "install",
            "-r",
            requirements_file,
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()



"""
GraphQL queries needed for pytoil's API calls.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

GET_REPO_NAMES = """
query ($username: String!, $limit: Int!) {
  user(login: $username) {
    repositories(first: $limit, ownerAffiliations: OWNER, orderBy: {field: NAME, direction: ASC}) {
      nodes {
        name
      }
    }
  }
}
"""

CHECK_REPO_EXISTS = """
query ($username: String!, $name: String!) {
  repository(owner: $username, name: $name) {
    name
  }
}
"""


GET_REPO_INFO = """
query ($username: String!, $name: String!) {
  repository(owner: $username, name: $name) {
    name,
    description,
    createdAt,
    updatedAt,
    diskUsage,
    licenseInfo {
      name
    }
  }
}
"""

GET_REPOS = """
query ($username: String!, $limit: Int!) {
  user(login: $username) {
    repositories(first: $limit, ownerAffiliations: OWNER, orderBy: {field: NAME, direction: ASC}) {
      nodes {
        name,
        description,
        createdAt,
        updatedAt,
        diskUsage
      }
    }
  }
}
"""

GET_FORKS = """
query ($username: String!, $limit: Int!) {
  user(login: $username) {
    repositories(first: $limit, ownerAffiliations: OWNER, isFork: true, orderBy: {field: NAME, direction: ASC}) {
      nodes {
        name
        diskUsage
        createdAt
        updatedAt
        parent {
          nameWithOwner
        }
      }
    }
  }
}
"""



"""
Module responsible for handling conda environments.


Author: Tom Fleet
Created: 26/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path
from typing import Sequence, Union

import aiofiles
import aiofiles.os
import yaml

from pytoil.exceptions import (
    BadEnvironmentFileError,
    CondaNotInstalledError,
    EnvironmentAlreadyExistsError,
    EnvironmentDoesNotExistError,
    UnsupportedCondaInstallationError,
)

# Type alias
EnvironmentYml = dict[str, Union[list[str], str]]

CONDA = shutil.which("conda")


class Conda:
    def __init__(
        self,
        root: Path,
        environment_name: str,
        conda: str | None = CONDA,
    ) -> None:
        """
        Representation of a Conda managed virtual environment.

        Conda environments are simpler in a lot of ways because
        it's all managed under the hood by the `conda` command.

        We don't need to worry about interpreters or paths etc.
        Just the environment name is enough to identify it.

        Args:
            root (Path): The root directory of the project.
            environment_name (str): The name of the conda environment
            conda (Optional[str]): The conda binary. Defaults to shutil.which("conda")
        """
        self.root = root
        self.environment_name = environment_name
        self.conda = conda

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(root={self.root!r}, environment_name={self.environment_name!r},"
            f" conda={self.conda!r})"
        )

    __slots__ = ("root", "environment_name", "conda")

    @property
    def project_path(self) -> Path:
        return self.root.resolve()

    @property
    def executable(self) -> Path:
        envs_dir = self.get_envs_dir()
        return envs_dir.joinpath(f"{self.environment_name}/bin/python")

    @property
    def name(self) -> str:
        return "conda"

    @staticmethod
    def get_envs_dir() -> Path:
        """
        Tries to detect which anaconda/miniconda installation
        the user has by naively checking if the environment
        storage directories exist.

        Raises:
            UnsupportedCondaInstallationError: If no matches are found
                for known conda installations.

        Returns:
            Path: The path to the users conda
                environment storage directory.
        """
        # Map of all supported conda installations to their root directories
        supported: dict[str, Path] = {
            "anaconda": Path.home().joinpath("anaconda3"),
            "miniconda": Path.home().joinpath("miniconda3"),
            "miniforge": Path.home().joinpath("miniforge3"),
            "mambaforge": Path.home().joinpath("mambaforge"),
        }

        names = supported.keys()

        for _, directory in supported.items():
            if directory.exists() and directory.is_dir():
                return directory.joinpath("envs")

        raise UnsupportedCondaInstallationError(
            f"""Could not detect the type of conda
        installation present. Checked for: {names}"""
        )

    async def exists(self) -> bool:
        """
        Checks whether the environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then the environment musty
        exist.
        """
        return await aiofiles.os.path.exists(self.executable)

    async def create(
        self, packages: Sequence[str] | None = None, silent: bool = False
    ) -> None:
        """
        Creates the conda environment described by the instance.

        Only default package is `python=3` which will cause conda to choose
        it's default python3.

        If packages are specified, these will be inserted into the conda create
        command.

        Args:
            packages (Optional[List[str]]): List of packages to install on creation.
                Defaults to None.

            silent (bool, optional): Whether to discard or display output.
                Defaults to False.

        Raises:
            CondaNotInstalledError: If `conda` not found on $PATH.
            EnvironmentAlreadyExistsError: If the conda environment already exists.
        """
        if not self.conda:
            raise CondaNotInstalledError

        if await self.exists():
            raise EnvironmentAlreadyExistsError(
                f"Conda env: {self.environment_name!r} already exists"
            )

        if packages:
            proc = await asyncio.create_subprocess_exec(
                self.conda,
                "create",
                "-y",
                "--name",
                self.environment_name,
                "python=3",
                *packages,
                cwd=self.project_path,
                stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
                stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
            )
        else:
            proc = await asyncio.create_subprocess_exec(
                self.conda,
                "create",
                "-y",
                "--name",
                self.environment_name,
                "python=3",
                cwd=self.project_path,
                stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
                stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
            )

        await proc.wait()

    @staticmethod
    async def create_from_yml(project_path: Path, silent: bool = False) -> None:
        """
        Creates a conda environment from the `environment.yml` contained
        in the root `project_path`

        Args:
            project_path (Path): Filepath to the project root.
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.

        Raises:
            BadEnvironmentFileError: If `environment.yml` is malformed.
            EnvironmentAlreadyExistsError: If a conda environment of the same name
                exists on the system.
        """

        if not shutil.which("conda"):
            raise CondaNotInstalledError

        # Ensure we have a fully resolved project path
        project_path = project_path.resolve()
        yml_file = project_path.joinpath("environment.yml")

        async with aiofiles.open(yml_file) as file:
            content = await file.read()
            env_dict: EnvironmentYml = yaml.full_load(content)

        env_name = env_dict.get("name")
        if not isinstance(env_name, str):
            raise BadEnvironmentFileError(
                "The environment yml file has an invalid format. Cannot determine value"
                " for key: `name`."
            )

        env = Conda(root=project_path, environment_name=env_name)

        if await env.exists():
            raise EnvironmentAlreadyExistsError(
                f"Conda env: {env_name!r} already exists."
            )

        # Can't use self.conda here as static method so just rely on $PATH
        proc = await asyncio.create_subprocess_exec(
            "conda",
            "env",
            "create",
            "--file",
            f"{yml_file}",
            cwd=project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def export_yml(self) -> None:
        """
        Exports an environment.yml file for the conda environment
        described by the instance.

        Raises:
            CondaNotInstalledError: If conda not installed.
            EnvironmentDoesNotExistError: If the environment does not exist.
        """
        if not self.conda:
            raise CondaNotInstalledError

        if not await self.exists():
            raise EnvironmentDoesNotExistError(
                f"Conda env: {self.environment_name!r} does not exist. Create it first"
                " before exporting the environment file."
            )

        proc = await asyncio.create_subprocess_exec(
            self.conda,
            "env",
            "export",
            "--from-history",
            "--name",
            self.environment_name,
            cwd=self.project_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        out, _ = await proc.communicate()

        yml_file = self.project_path.joinpath("environment.yml")

        async with aiofiles.open(yml_file, mode="w", encoding="utf-8") as file:
            await file.write(out.decode("utf-8"))

    async def install(self, packages: Sequence[str], silent: bool = False) -> None:
        """
        Installs `packages` into the conda environment described by
        the instance.

        Packages are passed straight through to conda so any versioning
        syntax will work as expected.

        Args:
            packages (List[str]): List of packages to install.
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.

        Raises:
            CondaNotInstalledError: If user does not have conda installed.
            EnvironmentDoesNotExistError: If the conda environment does not exist.
        """
        if not self.conda:
            raise CondaNotInstalledError

        if not await self.exists():
            raise EnvironmentDoesNotExistError(
                f"Conda env: {self.environment_name!r} does not exist. Create it first"
                " before installing packages."
            )

        proc = await asyncio.create_subprocess_exec(
            self.conda,
            "install",
            "-y",
            "--name",
            self.environment_name,
            *packages,
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def install_self(self, silent: bool = False) -> None:
        """
        Creates a conda environment from an environment.yml.

        This is conda's closest concept to `installing self`.

        Args:
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        await self.create_from_yml(project_path=self.project_path, silent=silent)



from pytoil.starters.go import GoStarter
from pytoil.starters.python import PythonStarter
from pytoil.starters.rust import RustStarter

__all__ = [
    "PythonStarter",
    "GoStarter",
    "RustStarter",
]



"""
The pytoil show command group.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncio
from datetime import datetime
from pathlib import Path
from typing import Any

import aiofiles.os
import asyncclick as click
import httpx
import humanize
from rich.console import Console
from rich.table import Table, box
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config

GITHUB_TIME_FORMAT = r"%Y-%m-%dT%H:%M:%SZ"


@click.group()
async def show() -> None:
    """
    View your local/remote projects.

    The show command provides an easy way of listing of the projects you have locally
    in your configured development directory and/or of those you have on GitHub
    (known in pytoil-land as 'remote' projects).

    Local projects will be the names of subdirectories in your configured projects
    directory.

    The remote projects listed here will be those owned by you on GitHub.

    The "--limit/-l" flag can be used if you only want to see a certain number
    of results.
    """


@show.command()
@click.option(
    "-l",
    "--limit",
    type=int,
    default=30,
    help="Maximum number of projects to list.",
    show_default=True,
)
@click.pass_obj
async def local(config: Config, limit: int) -> None:
    """
    Show your local projects.

    Show the projects you have locally in your configured
    projects directory.

    You can limit the number of projects shown with the
    "--limit/-l" flag.

    Examples:

    $ pytoil show local

    $ pytoil show local --limit 5
    """
    local_projects: set[Path] = {
        f
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    if not local_projects:
        msg.warn("You don't have any local projects yet!", exits=1)

    table = Table(box=box.SIMPLE)
    table.add_column("Name", style="bold white")
    table.add_column("Created")
    table.add_column("Modified")

    stats = await asyncio.gather(
        *[aiofiles.os.stat(project) for project in local_projects]
    )

    results = {project: stat for project, stat in zip(local_projects, stats)}

    click.secho("Local Projects", fg="cyan", bold=True)
    click.secho(
        f"\nShowing {min(limit, len(results))} out of {len(local_projects)} local"
        " projects",
        fg="bright_black",
        italic=True,
    )
    for path, result in sorted(results.items(), key=lambda x: str.casefold(str(x[0])))[
        :limit
    ]:
        table.add_row(
            path.name,
            humanize.naturaltime(datetime.utcfromtimestamp(result.st_birthtime)),
            humanize.naturaltime(datetime.utcfromtimestamp(result.st_mtime)),
        )

    console = Console()
    console.print(table)


@show.command()
@click.option(
    "-l",
    "--limit",
    type=int,
    default=30,
    help="Maximum number of projects to list.",
    show_default=True,
)
@click.pass_obj
async def remote(config: Config, limit: int) -> None:
    """
    Show your remote projects.

    Show the projects that you own on GitHub.

    These may include some you already have locally.
    Use 'show diff' to see the difference between local and remote.

    The "-l/--limit" flag can be used to limit the number of repos
    returned.

    Examples:

    $ pytoil show remote

    $ pytoil show remote --limit 10
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)

    try:
        repos = await api.get_repos()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not repos:
            msg.warn("You don't have any projects on GitHub yet.", exits=1)
            return

        table = Table(box=box.SIMPLE)
        table.add_column("Name", style="bold white")
        table.add_column("Size")
        table.add_column("Created")
        table.add_column("Modified")

        click.secho("Remote Projects", fg="cyan", bold=True)
        click.secho(
            f"\nShowing {min(limit, len(repos))} out of {len(repos)} remote projects",
            fg="bright_black",
            italic=True,
        )

        for repo in repos[:limit]:
            table.add_row(
                repo["name"],
                humanize.naturalsize(int(repo["diskUsage"]) * 1024),
                humanize.naturaltime(
                    datetime.strptime(repo["createdAt"], GITHUB_TIME_FORMAT)
                ),
                humanize.naturaltime(
                    datetime.strptime(repo["updatedAt"], GITHUB_TIME_FORMAT)
                ),
            )

        console = Console()
        console.print(table)


@show.command()
@click.option(
    "-l",
    "--limit",
    type=int,
    default=30,
    help="Maximum number of projects to list.",
    show_default=True,
)
@click.pass_obj
async def forks(config: Config, limit: int) -> None:
    """
    Show your forked projects.

    Show the projects you own on GitHub that are forks
    of other projects.

    The "-l/--limit" flag can be used to limit the number of
    repos returned.

    Examples:

    $ pytoil show forks

    $ pytoil show forks --limit 10
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)

    try:
        forks = await api.get_forks()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not forks:
            msg.warn("You don't have any forks yet.", exits=1)
            return

        table = Table(box=box.SIMPLE)
        table.add_column("Name", style="bold white")
        table.add_column("Size")
        table.add_column("Forked")
        table.add_column("Modified")
        table.add_column("Parent")

        click.secho("Forked Projects", fg="cyan", bold=True)
        click.secho(
            f"\nShowing {min(limit, len(forks))} out of {len(forks)} forked projects",
            fg="bright_black",
            italic=True,
        )

        for repo in forks[:limit]:
            table.add_row(
                repo["name"],
                humanize.naturalsize(int(repo["diskUsage"]) * 1024),
                humanize.naturaltime(
                    datetime.strptime(repo["createdAt"], GITHUB_TIME_FORMAT)
                ),
                humanize.naturaltime(
                    datetime.strptime(repo["updatedAt"], GITHUB_TIME_FORMAT)
                ),
                repo["parent"]["nameWithOwner"],
            )

        console = Console()
        console.print(table)


@show.command()
@click.option(
    "-l",
    "--limit",
    type=int,
    default=30,
    help="Maximum number of projects to list.",
    show_default=True,
)
@click.pass_obj
async def diff(config: Config, limit: int) -> None:
    """
    Show the difference in local/remote projects.

    Show the projects that you own on GitHub but do not
    have locally.

    The "-l/--limit" flag can be used to limit the number of repos
    returned.

    Examples:

    $ pytoil show diff

    $ pytoil show diff --limit 10
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)

    local_projects: set[str] = {
        f.name
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    try:
        remote_projects = await api.get_repos()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not remote_projects:
            msg.warn("You don't have any projects on GitHub yet!", exits=1)
            return

        remote_names: set[str] = {repo["name"] for repo in remote_projects}
        diff = remote_names.difference(local_projects)

        diff_info: list[dict[str, Any]] = []
        for repo in remote_projects:
            if name := repo.get("name"):
                if name in diff:
                    diff_info.append(repo)

        if not diff:
            msg.good("Your local and remote projects are in sync!")
        else:
            table = Table(box=box.SIMPLE)
            table.add_column("Name", style="bold white")
            table.add_column("Size")
            table.add_column("Created")
            table.add_column("Modified")

            click.secho("Diff: Remote - Local", fg="cyan", bold=True)
            click.secho(
                f"\nShowing {min(limit, len(diff_info))} out of"
                f" {len(diff_info)} projects",
                fg="bright_black",
                italic=True,
            )

            for repo in diff_info[:limit]:
                table.add_row(
                    repo["name"],
                    humanize.naturalsize(int(repo["diskUsage"] * 1024)),
                    humanize.naturaltime(
                        datetime.strptime(repo["createdAt"], GITHUB_TIME_FORMAT)
                    ),
                    humanize.naturaltime(
                        datetime.strptime(repo["updatedAt"], GITHUB_TIME_FORMAT)
                    ),
                )

            console = Console()
            console.print(table)



"""
Module responsible for interacting with local/remote projects.


Author: Tom Fleet
Created: 22/12/2021
"""

from __future__ import annotations

import asyncio
from datetime import datetime
from pathlib import Path
from typing import Any

import aiofiles
import aiofiles.os
import humanize
import tomlkit

from pytoil.api import API
from pytoil.environments import Conda, Environment, Flit, Poetry, Requirements, Venv
from pytoil.exceptions import RepoNotFoundError


class Repo:
    def __init__(self, owner: str, name: str, local_path: Path) -> None:
        """
        Representation of a local/remote project.

        Args:
            owner (str): The owner (i.e. GitHub username)
            name (str): Project name.
            local_path (Path): The path this repo would have on the
                local filesystem, whether it currently exists or not.
        """
        self.owner = owner
        self.name = name
        self.local_path = local_path

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(owner={self.owner!r}, name={self.name!r},"
            f" local_path={self.local_path!r})"
        )

    __slots__ = ("owner", "name", "local_path")

    @property
    def clone_url(self) -> str:
        """
        The url to feed to `git clone` in order to
        clone this repo.
        """
        return f"https://github.com/{self.owner}/{self.name}.git"

    @property
    def html_url(self) -> str:
        """
        The url to the homepage of this repo.
        """
        return f"https://github.com/{self.owner}/{self.name}"

    async def exists_local(self) -> bool:
        """
        Determines whether or not a `Repo` exists
        on the local filesystem.

        Returns:
            bool: True if exists locally, else False.
        """
        return await aiofiles.os.path.exists(self.local_path)

    async def exists_remote(self, api: API) -> bool:
        """
        Determines whether or not a `Repo` exists
        on GitHub.

        Args:
            api (API): The API object.

        Returns:
            bool: True if exists remote, else False.
        """
        return await api.check_repo_exists(self.name)

    async def _local_info(self) -> dict[str, Any] | None:
        """
        Return local path information for the repo.
        """
        try:
            return {
                "Name": self.local_path.name,
                "Created": humanize.naturaltime(
                    datetime.utcfromtimestamp(self.local_path.stat().st_birthtime)
                ),
                "Updated": humanize.naturaltime(
                    datetime.utcfromtimestamp(self.local_path.stat().st_mtime)
                ),
                "Local": True,
            }
        except FileNotFoundError:
            return None

    async def _remote_info(self, api: API) -> dict[str, Any] | None:
        """
        Return remote API information for the repo.
        """
        return await api.get_repo_info(self.name)

    async def info(self, api: API) -> dict[str, Any]:
        """
        Return summary info about the repo in question.

        Prefers the API info as it is possible to get
        things like license and description.

        If local only will fall back to path.stat.

        Args:
            api (API): The API dependency.

        Raises:
            RepoNotFoundError: If repo not local or remote.

        Returns:
            Dict[str, Any]: Repository information.
        """
        info: dict[str, Any] = {}
        exists_remote, remote_info, exists_local, local_info = await asyncio.gather(
            self.exists_remote(api=api),
            self._remote_info(api=api),
            self.exists_local(),
            self._local_info(),
        )

        if exists_remote:
            if remote_info:
                info.update(remote_info)
            # Might also exist locally
            info.update({"Local": exists_local})
        elif exists_local:
            if local_info:
                info.update(local_info)
            # We know it doesn't exist on GitHub if we got here
            info.update({"Remote": False})
        else:
            raise RepoNotFoundError(
                f"Repo: {self.name!r} does not exist locally or on GitHub."
            )

        return info

    async def _file_exists(self, file: str) -> bool:
        """
        Convenience method to determine whether or not
        the repo root directory contains `file`.

        Args:
            file (str): Name of the file.

        Returns:
            bool: True if exists, else False.
        """
        return await aiofiles.os.path.exists(self.local_path.joinpath(file))

    async def is_setuptools(self) -> bool:
        """
        Is the project based on setuptools, i.e. does it
        have a `setup.cfg` or `setup.py`

        Returns:
            bool: True if setuptools, else False.
        """
        exists = await asyncio.gather(
            self._file_exists("setup.cfg"), self._file_exists("setup.py")
        )

        return any(exists)

    async def is_requirements(self) -> bool:
        """
        Is the project a python app with a requirements file?

        Returns:
            bool: True if yes, else False.
        """
        exists = await asyncio.gather(
            self._file_exists("requirements.txt"),
            self._file_exists("requirements_dev.txt"),
        )

        return any(exists)

    async def has_pyproject_toml(self) -> bool:
        """
        Does the project have a `pyproject.toml` file.

        Returns:
            bool: True if yes, else False.
        """
        return await self._file_exists("pyproject.toml")

    async def is_conda(self) -> bool:
        """
        Is the project based on conda. The only reliable
        way of detecting this is looking for an `environment.yml`.

        Returns:
            bool: True if conda, else False.
        """
        return await self._file_exists("environment.yml")

    async def _specifies_build_tool(self, build_tool: str) -> bool:
        """
        Generalised method to check for a particular build tool
        in `pyproject.toml`

        Does more than a naive search for `build_tool`, actually checks the
        appropriate toml construction so if this method returns True, caller
        can be confident that `pyproject.toml` is valid.

        Args:
            build_tool (str): The build tool to check for e.g. `flit`, `poetry`.

        Returns:
            bool: True if `pyproject.toml` specifies that build tool, else False.
        """
        # If it doesn't have a pyproject.toml, bail early
        if not await self.has_pyproject_toml():
            return False

        async with aiofiles.open(self.local_path.joinpath("pyproject.toml")) as file:
            content = await file.read()
            toml = tomlkit.parse(content)

        if build_system := toml.get("build-system"):
            if build_backend := build_system.get("build-backend"):
                return build_tool in build_backend.strip().lower()

        return False

    async def is_poetry(self) -> bool:
        """
        Does the project specify a poetry build system

        Returns:
            bool: True if yes, else False.
        """
        return await self._specifies_build_tool("poetry")

    async def is_flit(self) -> bool:
        """
        Does the proeject specify a flit build system.

        Returns:
            bool: True if yes, else False.
        """
        return await self._specifies_build_tool("flit")

    async def dispatch_env(self) -> Environment | None:
        """
        Returns the correct environment object for the calling `Repo`,
        or `None` if it cannot detect the environment.

        Therefore all usage should first check for `None`.

        Returns:
            Optional[Environment]: The correct environment object if it was
                able to detect, or `None`.
        """
        # This is where the magic happens for automatic environment detection
        # and installation

        # Each of the environment objects below implements the `Environment` Protocol
        # and has an `install_self` method that does the correct thing for it's environment

        # Check the existence of files concurrently because why not
        exists = await asyncio.gather(
            self.is_conda(),
            self.is_requirements(),
            self.is_setuptools(),
            self.is_poetry(),
            self.is_flit(),
        )

        conda, requirements, setuptools, poetry, flit = exists

        if conda:
            return Conda(root=self.local_path, environment_name=self.name)
        elif requirements:
            return Requirements(root=self.local_path)
        elif setuptools:
            return Venv(root=self.local_path)
        elif poetry:
            return Poetry(root=self.local_path)
        elif flit:
            return Flit(root=self.local_path)
        else:
            # Could not autodetect, this is handled by the CLI
            return None



"""
The pytoil find command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncclick as click
from rich.console import Console
from rich.table import Table, box
from rich.text import Text
from thefuzz import process
from wasabi import msg

from pytoil.api import API
from pytoil.config import Config

FUZZY_SCORE_CUTOFF = 75


@click.command()
@click.argument("project", nargs=1)
@click.option(
    "-l",
    "--limit",
    type=int,
    default=3,
    help="Limit results to maximum number.",
    show_default=True,
)
@click.pass_obj
async def find(config: Config, project: str, limit: int) -> None:
    """
    Quickly locate a project.

    The find command provides a fuzzy search for finding a project when you
    don't know where it is (local or on GitHub).

    It will perform a fuzzy search through all your local and remote projects,
    bring back the best matches and show you where they are.

    Useful if you have a lot of projects and you can't quite remember
    what the one you want is called!

    The "-l/--limit" flag can be used to alter the number of returned
    search results, but bare in mind that matches with sufficient match score
    are returned anyway so the results flag only limits the maximum number
    of results shown.

    Examples:

    $ pytoil find my

    $ pytoil find proj --limit 5
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)

    local_projects: set[str] = {
        f.name
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }
    remote_projects = await api.get_repo_names()

    all_projects = local_projects.union(remote_projects)

    matches: list[tuple[str, int]] = process.extractBests(
        project, all_projects, limit=limit, score_cutoff=FUZZY_SCORE_CUTOFF
    )

    table = Table(box=box.SIMPLE)
    table.add_column("Project", style="bold white")
    table.add_column("Similarity")
    table.add_column("Where")

    if len(matches) == 0:
        msg.warn("No matches found!", exits=1)

    for match in matches:
        is_local = match[0] in local_projects
        table.add_row(
            match[0],
            str(match[1]),
            Text("Local", style="green")
            if is_local
            else Text("Remote", style="dark_orange"),
        )

    console = Console()
    console.print(table)



"""
Collection of useful helpers for the CLI.


Author: Tom Fleet
Created: 30/12/2021
"""
from __future__ import annotations

import httpx
from wasabi import msg


def handle_http_status_error(error: httpx.HTTPStatusError) -> None:
    """
    Handles a variety of possible HTTP Status errors, print's nicer output
    to the user, and exits the program if necessary.
    Call this in an except block on CLI commands accessing the
    GitHub API.
    Args:
        error (httpx.HTTPStatusError): The error to be handled.
    """
    code = error.response.status_code

    if code == 401:
        msg.fail(
            title="HTTP Error: 401 - Unauthorized",
            text="This usually means something is wrong with your token!",
            exits=1,
        )
    elif code == 404:
        msg.fail(
            title="HTTP Error: 404 - Not Found",
            text="This usually means you've made a typo.",
            exits=1,
        )
    elif code == 500:
        msg.fail(
            title="HTTP Error: 500 - Internal Server Error",
            text="This usually means GitHub is not happy.",
            exits=1,
        )



"""
The python starter template.


Author: Tom Fleet
Created: 29/12/2021
"""


from __future__ import annotations

from pathlib import Path

import aiofiles
import aiofiles.os


class PythonStarter:
    def __init__(self, path: Path, name: str) -> None:
        self.path = path
        self.name = name
        self.root = self.path.joinpath(self.name).resolve()
        self.files = [
            self.root.joinpath(filename)
            for filename in ["README.md", "requirements.txt", f"{self.name}.py"]
        ]

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(path={self.path!r}, name={self.name!r})"

    __slots__ = ("path", "name", "root", "files")

    async def generate(self, username: str | None = None) -> None:
        """
        Generate a new python starter template.
        """
        await aiofiles.os.mkdir(self.root)

        for file in self.files:
            file.touch()

        # Put the header in the README
        readme = self.root.joinpath("README.md")
        async with aiofiles.open(readme, mode="w", encoding="utf-8") as f:
            await f.write(f"# {self.name}\n")

        # Put a hint in the requirements.txt
        reqs = self.root.joinpath("requirements.txt")
        async with aiofiles.open(reqs, mode="w", encoding="utf-8") as f:
            await f.write("# Put your requirements here e.g. flask>=1.0.0\n")

        # Populate the python file
        py_file = self.root.joinpath(f"{self.name}.py")
        py_text = (
            'def hello(name: str = "world") -> None:\n    print(f"hello {name}")\n'
        )
        async with aiofiles.open(py_file, mode="w", encoding="utf-8") as f:
            await f.write(py_text)



"""
The pytoil gh command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncclick as click
import httpx
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.repo import Repo


@click.command()
@click.argument("project", nargs=1)
@click.pass_obj
async def gh(config: Config, project: str) -> None:
    """
    Open one of your projects on GitHub.

    Given a project name (must exist on GitHub and be owned by you),
    'gh' will open your browser and navigate to the project on GitHub.

    Examples:

    $ pytoil gh my_project
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        exists = await repo.exists_remote(api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not exists:
            msg.warn(f"Could not find {project!r} on GitHub. Was it a typo?", exits=1)
        else:
            msg.info(f"Opening {project!r} on GitHub")
            click.launch(url=repo.html_url)



"""
The pytoil new command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncio
import functools

import aiofiles.os
import asyncclick as click
from cookiecutter.main import cookiecutter
from wasabi import msg

from pytoil.api import API
from pytoil.config import Config
from pytoil.environments import Conda, Venv
from pytoil.exceptions import (
    CargoNotInstalledError,
    EnvironmentAlreadyExistsError,
    GoNotInstalledError,
)
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.starters import GoStarter, PythonStarter, RustStarter
from pytoil.vscode import VSCode


@click.command()
@click.argument("project", nargs=1, type=str)
@click.argument("packages", nargs=-1)
@click.option(
    "-c",
    "--cookie",
    type=str,
    help="URL to a cookiecutter template from which to build the project.",
)
@click.option(
    "-s",
    "--starter",
    type=click.Choice(choices={"python", "go", "rust"}, case_sensitive=True),
    help="Use a language-specific starter template.",
)
@click.option(
    "-v",
    "--venv",
    type=click.Choice(choices={"venv", "conda"}, case_sensitive=True),
    help="Which type of virtual environment to create.",
)
@click.option(
    "-n",
    "--no-git",
    is_flag=True,
    default=False,
    help="Don't do any git stuff.",
)
@click.pass_obj
async def new(  # noqa: C901
    config: Config,
    project: str,
    packages: tuple[str, ...],
    cookie: str | None,
    starter: str | None,
    venv: str | None,
    no_git: bool = False,
) -> None:
    """
    Create a new development project.

    Bare usage will simply create an empty folder in your configured projects
    directory.

    You can also create a project from a cookiecutter template by passing a valid
    url to the '--cookie/-c' flag.

    If you just want a very simple, language-specific starting template, use the
    '--starter/-s' option.

    By default, pytoil will initialise a local git repo in the folder and commit it,
    following the style of modern language build tools such as rust's cargo. You can disable
    this behaviour by setting 'git' to false in pytoil's config file
    or by passing the '--no-git/-n' flag here.

    If you want pytoil to create a new virtual environment for your project, you
    can use the '--venv/-v' flag. Standard python and conda virtual environments
    are supported.

    If the '--venv/-v' flag is used, you may also pass a list of python packages
    to install into the created virtual environment. These will be delegated to
    the appropriate tool (pip or conda) depending on what environment was created.
    If the environment is conda, the packages will be passed at environment creation
    time meaning they will have their dependencies resolved together. Normal python
    environments will first be created and then have specified packages installed.

    If 'common_packages' is specified in pytoil's config file, these will automatically
    be included in the environment.

    To specify versions of packages via the command line, you must enclose them
    in double quotes e.g. "flask>=1.0.0" not flask>=1.0.0 otherwise this will
    be interpreted by the shell as a command redirection.

    Examples:

    $ pytoil new my_project

    $ pytoil new my_project --cookie https://github.com/some/cookie.git

    $ pytoil new my_project --venv conda

    $ pytoil new my_project -c https://github.com/some/cookie.git -v conda --no-git

    $ pytoil new my_project -v venv requests "flask>=1.0.0"

    $ pytoil new my_project --starter python
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path, code=config.code_bin)
    git = Git()

    # Additional packages to include
    to_install: list[str] = [*packages] + config.common_packages

    # Can't use --cookie and --starter
    if cookie and starter:
        msg.warn("--cookie and --starter are mutually exclusive", exits=1)

    # Can't use --venv with non-python starters
    if (
        starter is not None  # User specified --starter
        and starter != "python"  # Requested starter is not python
        and venv is not None  # And the user wants a virtual environment
    ):
        msg.warn(f"Can't create a venv for a {starter} project", exits=1)

    # Resolve config vs flag for no-git
    # flag takes priority over config
    use_git: bool = config.git and not no_git

    # Does this project already exist?
    # Mightaswell check concurrently
    local, remote = await asyncio.gather(repo.exists_local(), repo.exists_remote(api))

    if local:
        msg.warn(
            title=f"{repo.name} already exists locally.",
            text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
            exits=1,
        )

    if remote:
        msg.warn(
            title=f"{repo.name} already exists on GitHub.",
            text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
            exits=1,
        )

    # If we get here, we're good to create a new project
    if cookie:
        msg.info(f"Creating {repo.name!r} from cookiecutter: {cookie!r}.")
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(
            executor=None,
            func=functools.partial(
                cookiecutter,
                template=cookie,
                output_dir=config.projects_dir,
            ),
        )

    elif starter == "go":
        msg.info(f"Creating {repo.name!r} from starter: {starter!r}.")
        go_starter = GoStarter(path=config.projects_dir, name=repo.name)

        try:
            await go_starter.generate(username=config.username)
        except GoNotInstalledError:
            msg.fail("Error: Go not installed.", exits=1)
        else:
            if use_git:
                await git.init(cwd=repo.local_path, silent=False)
                await git.add(cwd=repo.local_path, silent=False)
                await git.commit(cwd=repo.local_path, silent=False)

    elif starter == "python":
        msg.info(f"Creating {repo.name!r} from starter: {starter!r}.")
        python_starter = PythonStarter(path=config.projects_dir, name=repo.name)
        await python_starter.generate()
        if use_git:
            await git.init(cwd=repo.local_path, silent=False)
            await git.add(cwd=repo.local_path, silent=False)
            await git.commit(cwd=repo.local_path, silent=False)

    elif starter == "rust":
        msg.info(f"Creating {repo.name!r} from starter: {starter!r}.")
        rust_starter = RustStarter(path=config.projects_dir, name=repo.name)

        try:
            await rust_starter.generate()
        except CargoNotInstalledError:
            msg.fail("Error: Cargo not installed.", exits=1)
        else:
            if use_git:
                await git.init(cwd=repo.local_path, silent=False)
                await git.add(cwd=repo.local_path, silent=False)
                await git.commit(cwd=repo.local_path, silent=False)

    else:
        # Just a blank new project
        msg.info(f"Creating {repo.name!r} at {repo.local_path}.")
        await aiofiles.os.mkdir(repo.local_path)
        if use_git:
            await git.init(cwd=repo.local_path, silent=False)

    # Now we need to handle any requested virtual environments
    if venv == "venv":
        msg.info(
            title=f"Creating virtual environment for {repo.name!r}",
            text=f"Including packages: {', '.join(to_install)}" if to_install else "",
        )
        env = Venv(root=repo.local_path)
        with msg.loading("Working..."):
            await env.create(packages=to_install, silent=True)

    elif venv == "conda":
        msg.info(
            title=f"Creating conda environment for {repo.name!r}",
            text=f"Including packages: {', '.join(to_install)}" if to_install else "",
        )
        conda_env = Conda(root=repo.local_path, environment_name=repo.name)
        try:
            await conda_env.create(packages=to_install)
        except EnvironmentAlreadyExistsError:
            msg.fail(
                f"Conda environment: {conda_env.environment_name!r} already exists",
                exits=1,
            )
        else:
            # Export the environment.yml
            await conda_env.export_yml()

    # Now handle opening in VSCode
    if config.vscode:
        msg.info(f"Opening {repo.name!r} in VSCode.")
        await code.open()



"""
The root CLI command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

from pathlib import Path

import aiofiles.os
import asyncclick as click
import questionary
from wasabi import msg

from pytoil import __version__
from pytoil.cli import (
    cache,
    checkout,
    config,
    docs,
    find,
    gh,
    info,
    new,
    pull,
    remove,
    show,
)
from pytoil.config import Config, defaults


@click.group(
    commands=(
        checkout.checkout,
        config.config,
        docs.docs,
        find.find,
        gh.gh,
        info.info,
        new.new,
        pull.pull,
        remove.remove,
        show.show,
        cache.cache,
    )
)
@click.version_option(version=__version__, package_name="pytoil", prog_name="pytoil")
@click.pass_context
async def main(ctx: click.Context) -> None:
    """
    Helpful CLI to automate the development workflow.

    - Create and manage your local and remote projects

    - Build projects from cookiecutter templates.

    - Easily create/manage virtual environments.

    - Minimal configuration required.
    """
    # Load the config once on launch of the app and pass it down to the child commands
    # through click's context
    try:
        config = await Config.load()
    except FileNotFoundError:
        msg.warn("No pytoil config file detected!")
        interactive: bool = await questionary.confirm(
            "Interactively configure pytoil?", default=False, auto_enter=False
        ).ask_async()

        if not interactive:
            # User doesn't want to interactively walk through a config file
            # so just make a default and exit cleanly
            await Config.helper().write()
            msg.good(
                title="I made a default file for you",
                text=f"It's here: {defaults.CONFIG_FILE}",
                spaced=True,
            )
            msg.text("Tip: You can edit it with `pytoil config edit`.", exits=0)
            return

        # If we get here, the user wants to interactively make the config
        projects_dir: str = await questionary.path(
            "Where do you keep your projects?",
            default=str(defaults.PROJECTS_DIR),
            only_directories=True,
        ).ask_async()

        token: str = await questionary.text("GitHub personal access token?").ask_async()

        username: str = await questionary.text(
            "What's your GitHub username?"
        ).ask_async()

        vscode: bool = await questionary.confirm(
            "Do you use VSCode?",
            default=False,
            auto_enter=False,
        ).ask_async()

        code_bin = defaults.CODE_BIN

        if vscode:

            which_vscode: str = await questionary.select(
                "Which version of VSCode do you use?",
                choices=["stable", "insiders"],
                default="stable",
            ).ask_async()

            code_bin = (
                "code-insiders" if which_vscode == "insiders" else defaults.CODE_BIN
            )

        git: bool = await questionary.confirm(
            "Make git repos when creating new projects?", default=True, auto_enter=False
        ).ask_async()

        config = Config(
            projects_dir=Path(projects_dir),
            token=token,
            username=username,
            vscode=vscode,
            code_bin=code_bin,
            git=git,
        )

        await config.write()

        msg.good(
            "Config created", text=f"It's available at {defaults.CONFIG_FILE}.", exits=0
        )
        return

    else:
        # We have a valid config file at the right place so load it into click's
        # context and pass it down to all subcommands
        ctx.obj = config

        # Ensure the API cache dir exists
        if not await aiofiles.os.path.exists(defaults.CACHE_DIR):
            await aiofiles.os.makedirs(defaults.CACHE_DIR)



"""
The pytoil docs command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncclick as click
from wasabi import msg

from pytoil.config import defaults


@click.command()
async def docs() -> None:
    """
    Open pytoil's documentation in your browser.

    Examples:

    $ pytoil docs
    """
    msg.info("Opening pytoil's docs in your browser...")
    click.launch(url=defaults.PYTOIL_DOCS_URL)



"""
Interface that all starter classes must satisfy.

These templates are not supposed to be exhaustive, for that the user
is better off using pytoil's cookiecutter functionality.

The templates defined in the `starters` module are exactly that,
a starter. A good analogous reference would be the behaviour of
`cargo new` in rust, which simply sets up a few basic sub directories
and a "hello world" function main.rs.


Author: Tom Fleet
Created: 29/12/2021
"""

from __future__ import annotations

from typing import Protocol


class Starter(Protocol):
    async def generate(self, username: str | None = None) -> None:
        """
        Implements the generation of the project starter template.
        """
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytoil',
 'pytoil.api',
 'pytoil.cli',
 'pytoil.config',
 'pytoil.environments',
 'pytoil.git',
 'pytoil.repo',
 'pytoil.starters',
 'pytoil.vscode']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML==6.0',
 'aiofiles==0.8.0',
 'anyio==3.5.0',
 'asyncclick==8.0.3.2',
 'attrs==21.4.0',
 'cookiecutter==1.7.3',
 'httpx-cache==0.4.0',
 'httpx[http2]==0.21.3',
 'humanize==3.13.1',
 'pydantic==1.9.0',
 'questionary==1.10.0',
 'rich==11.0.0',
 'thefuzz[speedup]==0.19.0',
 'tomlkit==0.8.0',
 'virtualenv==20.13.0',
 'wasabi==0.9.0']

entry_points = \
{'console_scripts': ['pytoil = pytoil.cli.root:main']}

setup_kwargs = {
    'name': 'pytoil',
    'version': '0.23.2',
    'description': 'CLI to automate the development workflow.',
    'long_description': "![logo](https://github.com/FollowTheProcess/pytoil/raw/main/docs/img/logo.png)\n\n[![License](https://img.shields.io/github/license/FollowTheProcess/pytoil)](https://github.com/FollowTheProcess/pytoil)\n[![PyPI](https://img.shields.io/pypi/v/pytoil.svg?logo=python)](https://pypi.python.org/pypi/pytoil)\n[![GitHub](https://img.shields.io/github/v/release/FollowTheProcess/pytoil?logo=github&sort=semver)](https://github.com/FollowTheProcess/pytoil)\n[![Code Style](https://img.shields.io/badge/code%20style-black-black)](https://github.com/FollowTheProcess/pytoil)\n[![CI](https://github.com/FollowTheProcess/pytoil/workflows/CI/badge.svg)](https://github.com/FollowTheProcess/pytoil/actions?query=workflow%3ACI)\n[![Coverage](https://github.com/FollowTheProcess/pytoil/raw/main/docs/img/coverage.svg)](https://github.com/FollowTheProcess/pytoil)\n\n*pytoil is a small, helpful CLI to help developers manage their local and remote projects with ease!*\n\n* **Source Code**: [https://github.com/FollowTheProcess/pytoil](https://github.com/FollowTheProcess/pytoil)\n\n* **Documentation**: [https://FollowTheProcess.github.io/pytoil/](https://FollowTheProcess.github.io/pytoil/)\n\n## What is it?\n\n`pytoil` is a handy tool that helps you stay on top of all your projects, remote or local. It's primarily aimed at python developers but you could easily use it to manage any project!\n\npytoil is:\n\n* Easy to use ✅\n* Easy to configure ✅\n* Safe (it won't edit your repos at all) ✅\n* Snappy (it's asynchronous from the ground up and as much as possible is done concurrently, clone all your repos in seconds!) 💨\n* Useful! (I hope 😃)\n\nSay goodbye to janky bash scripts 👋🏻\n\n## Background\n\nLike many developers I suspect, I quickly became bored of typing repeated commands to manage my projects, create virtual environments, install packages, fire off `cURL` snippets to check if I had a certain repo etc.\n\nSo I wrote some shell functions to do some of this for me...\n\nAnd these shell functions grew and grew and grew.\n\nUntil one day I saw that the file I kept these functions in was over 1000 lines of bash (a lot of `printf`'s so it wasn't all logic but still). And 1000 lines of bash is *waaaay* too much!\n\nAnd because I'd basically hacked it all together, it was **very** fragile. If a part of a function failed, it would just carry on and wreak havoc! I'd have to do `rm -rf all_my_projects`... I mean careful forensic investigation to fix it.\n\nSo I decided to make a robust CLI with the proper error handling and testability of python, and here it is! 🎉\n\n## Installation\n\nAs pytoil is a CLI program, I'd recommend installing with [pipx].\n\n```shell\n$ pipx install pytoil\n---> 100%\nSuccessfully installed pytoil\n```\n\nYou can always fall back to pip\n\n```shell\n$ python3 -m pip install pytoil\n---> 100%\nSuccessfully installed pytoil\n```\n\npytoil will install everything it needs *in python* to work. However, it's full feature set can only be accessed if you have the following external dependencies:\n\n* [git]\n* [conda] (if you work with conda environments)\n* [VSCode] (if you want to use pytoil to automatically open your projects for you)\n* [poetry] (if you want to create poetry environments)\n* [flit] (if you want to create flit environments)\n\n## Quickstart\n\n`pytoil` is super easy to get started with.\n\nAfter you install pytoil, the first time you run it you'll get something like this.\n\n```plain\n$ pytoil <any command>\n\nNo pytoil config file detected!\n? Interactively configure pytoil? [y/n]\n```\n\nIf you say yes, pytoil will walk you through a few questions and fill out your config file with the values you enter. If you'd rather not do this interactively, just say no and it will instead put a default config file in the right place for you to edit later.\n\nOnce you've configured it properly, you can do things like...\n\n#### See your local and remote projects\n\n```plain\n$ pytoil show local\nLocal Projects\n\nShowing 3 out of 3 local projects\n\n  Name              Created          Modified\n ───────────────────────────────────────────────────\n  project 1         13 days ago      9 days ago\n  project 2         a day ago        a minute ago\n  project 3         a month ago      a month ago\n```\n\n#### See which ones you have on GitHub, but not on your computer\n\n```plain\n$ pytoil show diff\nDiff: Remote - Local\n\nShowing 3 out of 3 projects\n\n  Name             Size       Created                Modified\n ─────────────────────────────────────────────────────────────────────────\n  remote 1         154.6 kB   a month ago            29 days ago\n  remote 2         2.1 MB     1 year, 15 days ago    11 months ago\n  remote 3         753.7 kB   1 year, 6 months ago   a month ago\n```\n\n#### Easily grab a project, regardless of where it is\n\n```plain\n$ pytoil checkout myproject\n\n// Will now either open that project if local\n// or clone it, then open it if not\n```\n\n#### Create a new project and virtual environment in one go\n\n```plain\n$ pytoil new myproject --venv venv\n\nCreating project: 'myproject' at '/Users/you/projects/myproject'\n\nCreating virtual environment for: 'myproject'\n```\n\n#### And even do this from a [cookiecutter] template\n\n```plain\n$ pytoil new myproject --venv venv --cookie https://github.com/some/cookie.git\n\nCreating project: 'myproject' with cookiecutter template: 'https://github.com/some/cookie.git'\n\n// You'll then be asked all the cookiecutter questions defined in the template\n// After which pytoil will take over and create the virtual environment as normal\n```\n\nAnd loads more!\n\n### Help\n\nLike all good CLI programs, pytoil (as well as all it's subcommands, and all *their* subcommands!) has a `--help` option to show you what to do.\n\n```plain\n$ pytoil --help\n\nUsage: pytoil [OPTIONS] COMMAND [ARGS]...\n\n  Helpful CLI to automate the development workflow.\n\n  - Create and manage your local and remote projects\n\n  - Build projects from cookiecutter templates.\n\n  - Easily create/manage virtual environments.\n\n  - Minimal configuration required.\n\nOptions:\n  --version  Show the version and exit.\n  --help     Show this message and exit.\n\nCommands:\n  checkout  Checkout an existing development project.\n  config    Interact with pytoil's configuration.\n  docs      Open pytoil's documentation in your browser.\n  find      Quickly locate a project.\n  gh        Open one of your projects on GitHub.\n  info      Get useful info for a project.\n  new       Create a new development project.\n  pull      Pull down your remote projects.\n  remove    Remove projects from your local filesystem.\n  show      View your local/remote projects.\n```\n\npytoil's CLI is designed such that if you don't specify any arguments, it won't do anything! just show you the `--help`. This is called being a 'well behaved' unix command line tool.\n\nThis is true for any subcommand of pytoil so you won't accidentally break anything if you don't specify arguments 🎉\n\nAnd if you get truly stuck, you can quickly open pytoil's documentation with:\n\n```plain\n$ pytoil docs\n\nOpening pytoil's documentation in your browser...\n\n# Now you'll be on this page in whatever your default browser is!\n```\n\nCheck out the [docs] for more 💥\n\n## Contributing\n\n`pytoil` is an open source project and, as such, welcomes contributions of all kinds 😃\n\nYour best bet is to check out the [contributing guide] in the docs!\n\n[pipx]: https://pipxproject.github.io/pipx/\n[docs]: https://FollowTheProcess.github.io/pytoil/\n[FollowTheProcess/poetry_pypackage]: https://github.com/FollowTheProcess/poetry_pypackage\n[wasabi]: https://github.com/ines/wasabi\n[httpx]: https://www.python-httpx.org\n[async-click]: https://github.com/python-trio/asyncclick\n[contributing guide]: https://followtheprocess.github.io/pytoil/contributing/contributing.html\n[git]: https://git-scm.com\n[conda]: https://docs.conda.io/en/latest/\n[VSCode]: https://code.visualstudio.com\n[config]: config.md\n[cookiecutter]: https://github.com/cookiecutter/cookiecutter\n[poetry]: https://python-poetry.org\n[flit]: https://flit.readthedocs.io\n",
    'author': 'Tom Fleet',
    'author_email': 'tomfleet2018@gmail.com',
    'maintainer': 'Tom Fleet',
    'maintainer_email': 'tomfleet2018@gmail.com',
    'url': 'https://github.com/FollowTheProcess/pytoil',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
Module responsible for handling Flit python environments.


Author: Tom Fleet
Created: 26/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path

from pytoil.environments.virtualenv import Venv
from pytoil.exceptions import FlitNotInstalledError

FLIT = shutil.which("flit")


class Flit(Venv):
    def __init__(self, root: Path, flit: str | None = FLIT) -> None:
        self.root = root
        self.flit = flit
        super().__init__(root)

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r}, flit={self.flit!r})"

    __slots__ = ("root", "flit")

    @property
    def name(self) -> str:
        return "flit"

    async def install_self(self, silent: bool = False) -> None:
        """
        Installs a flit based project.

        Args:
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        if not self.flit:
            raise FlitNotInstalledError

        # Unlike poetry, conda etc. flit does not make it's own virtual environment
        # we must make one here before installing the project
        if not await self.exists():
            await self.create()

        proc = await asyncio.create_subprocess_exec(
            self.flit,
            "install",
            "--deps",
            "develop",
            "--symlink",
            "--python",
            f"{self.executable}",
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()



"""
The pytoil cache subcommand.


Author: Tom Fleet
Created: 21/01/2022
"""

from __future__ import annotations

import asyncio
import functools
import shutil

import asyncclick as click
from wasabi import msg

from pytoil.api import API
from pytoil.config import Config, defaults


@click.group()
async def cache() -> None:
    """
    Interact with pytoil's cache.

    Pytoil caches the responses from the GitHub API so that it
    doesn't have to hit it so often and so a user gets the snappiest
    experience possible using the CLI.

    The cache subcommand allows you to interact with this cache, either
    forcing a manual refresh or flushing it entirely.
    """


@cache.command()
@click.pass_obj
async def refresh(config: Config) -> None:
    """
    Force a manual refresh of the cache.

    This will cause pytoil to hit all the GitHub API endpoints
    it uses and cache the responses so that the next time any
    API calls are made, the responses will be served from the cache
    and not the GitHub API.

    Examples:

    $ pytoil cache refresh
    """
    api = API(username=config.username, token=config.token)
    names = await api.get_repo_names()

    # Just hit all the cacheable endpoints to initialise the cache
    await asyncio.gather(
        api.get_repos(),
        api.get_forks(),
        api.get_repo_names(),
        *[api.get_repo_info(name) for name in names]
    )

    msg.good("Cache refreshed successfully")


@cache.command()
@click.pass_obj
async def clear(config: Config) -> None:
    """
    Clear pytoil's cache.

    This will remove the entire cache forcing the next command
    using the GitHub API to fetch the latest data.

    It may be slightly slower as a result.

    Examples:

    $ pytoil cache clear
    """
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(
        executor=None,
        func=functools.partial(
            shutil.rmtree,
            path=defaults.CACHE_DIR,
            ignore_errors=True,
        ),
    )

    msg.good("Cache cleared successfully")



"""
The pytoil pull command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncio

import asyncclick as click
import httpx
import questionary
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.git import Git
from pytoil.repo import Repo


@click.command()
@click.argument("projects", nargs=-1)
@click.option("-f", "--force", is_flag=True, help="Force pull without confirmation.")
@click.option("-a", "--all", "all_", is_flag=True, help="Pull down all your projects.")
@click.pass_obj
async def pull(
    config: Config, projects: tuple[str, ...], force: bool, all_: bool
) -> None:
    """
    Pull down your remote projects.

    The pull command provides easy methods for pulling down remote projects.

    It is effectively a nice wrapper around git clone but you don't have to
    worry about urls or what your cwd is, pull will grab your remote projects
    by name and clone them to your configured projects directory.

    You can also use pull to batch clone multiple repos, even all of them ("--all/-a")
    if you're into that sorta thing.

    If more than 1 repo is passed (or if "--all/-a" is used) pytoil will pull
    the repos concurrently, speeding up the process.

    Any remote project that already exists locally will be skipped and none of
    your local projects are changed in any way. pytoil will only pull down
    those projects that don't already exist locally.

    It's very possible to accidentally clone a lot of repos when using pull so
    you will be prompted for confirmation before pytoil does anything.

    The "--force/-f" flag can be used to override this confirmation prompt if
    desired.

    Examples:

    $ pytoil pull project1 project2 project3

    $ pytoil pull project1 project2 project3 --force

    $ pytoil pull --all

    $ pytoil pull --all --force
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    if not projects and not all_:
        msg.warn(
            "If not using the '--all' flag, you must specify projects to pull.", exits=1
        )

    api = API(username=config.username, token=config.token)

    local_projects: set[str] = {
        f.name
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    try:
        remote_projects = await api.get_repo_names()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not remote_projects:
            msg.warn("You don't have any remote projects to pull.", exits=1)

        not_found = False
        specified_remotes = remote_projects if all_ else set(projects)

        # Check for typos
        for project in projects:
            if project not in remote_projects:
                not_found = True
                missing_project = project
                break
            break

        if not_found:
            msg.warn(
                f"{missing_project!r} not found on GitHub. Was it a typo?", exits=1
            )

        diff = specified_remotes.difference(local_projects)
        if not diff:
            msg.good("Your local and remote projects are in sync!", exits=0)

        if not force:
            if len(diff) <= 3:
                message = f"This will pull down {', '.join(diff)}. Are you sure?"
            else:
                # Too many to show nicely
                message = f"This will pull down {len(diff)} projects. Are you sure?"

            confirmed: bool = await questionary.confirm(
                message, default=False, auto_enter=False
            ).ask_async()

            if not confirmed:
                raise click.Abort()

        # Now we're good to go
        to_clone = [
            Repo(
                owner=config.username,
                name=project,
                local_path=config.projects_dir.joinpath(project),
            )
            for project in diff
        ]
        git = Git()
        await asyncio.gather(
            *[clone_and_report(repo, git, config) for repo in to_clone]
        )


async def clone_and_report(repo: Repo, git: Git, config: Config) -> None:
    await git.clone(url=repo.clone_url, cwd=config.projects_dir)
    msg.good(f"Cloned {repo.name!r}")



"""
The async-pytoil checkout command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import asyncio
import re

import asyncclick as click
import httpx
import questionary
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.environments import Environment
from pytoil.exceptions import (
    EnvironmentAlreadyExistsError,
    ExternalToolNotInstalledError,
)
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.vscode import VSCode

# The username/repo pattern
USER_REPO_REGEX = re.compile(r"^([A-Za-z0-9_.-])+/([A-Za-z0-9_.-])+$")

# The bare 'project' pattern
PROJECT_REGEX = re.compile(r"^([A-Za-z0-9_.-])+$")


@click.command()
@click.argument("project", nargs=1)
@click.option(
    "-v",
    "--venv",
    is_flag=True,
    help="Attempt to auto-create a virtual environment.",
)
@click.pass_obj
async def checkout(config: Config, project: str, venv: bool) -> None:
    """
    Checkout an existing development project.

    The checkout command lets you easily resume work on an existing project,
    whether that project is available locally in your configured projects
    directory, or if it is on GitHub.

    If pytoil finds your project locally, and you have enabled VSCode in your
    config file it will open it for you. If not, it will just tell you it already
    exists locally and where to find it.

    If your project is on your GitHub, pytoil will clone it for you and then open it
    (or tell you where it cloned it if you dont have VSCode set up).

    Finally, if checkout can't find a match after searching locally and on GitHub,
    it will prompt you to use 'pytoil new' to create a new one.

    If you pass the shorthand to someone elses repo e.g. 'someoneelse/repo' pytoil
    will detect this and automatically create a fork of this repo for you. Forking
    happens asynchronously so we give it a few seconds, then check whether or not
    your fork exists yet. If it does, all is well and we can clone it for you
    automatically. If not, (which is totally normal), we'll let you know. In
    which case just give it a few seconds then a 'pytoil checkout repo'
    will bring it down as normal.

    You can also ask pytoil to automatically create a virtual environment on
    checkout with the '--venv/-v' flag. This only happens for projects pulled down
    from GitHub to avoid accidentally screwing up local projects.

    If the '--venv/-v' flag is used, pytoil will look at your project to try and detect
    which type of environment to create e.g. conda, flit, poetry, standard python etc.

    The '--venv/-v' flag will also attempt to detect if the project you're checking out
    is a python package, in which case it will install it's requirements into the
    created environment.

    More info about this can be found in the documentation. Use `pytoil docs` to go there.

    Examples:

    $ pytoil checkout my_project

    $ pytoil checkout my_project --venv

    $ pytoil checkout someoneelse/project
    """
    if not config.can_use_api():
        msg.warn(
            "You must set your GitHub username and personal access token to use API"
            " features.",
            exits=1,
        )

    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    if bool(USER_REPO_REGEX.match(project)):
        # We've matched the "user/repo" pattern, meaning the user wants
        # to create a fork
        owner, name = project.split("/")
        if owner == config.username:
            msg.warn("You don't need the '/' when checking out a repo you own", exits=1)

        await checkout_fork(
            owner=owner,
            name=name,
            api=api,
            config=config,
            git=git,
            venv=venv,
        )

        msg.good("Done!")

    elif bool(PROJECT_REGEX.match(project)):

        if await repo.exists_local():
            await checkout_local(repo=repo, code=code, config=config, venv=venv)
        elif await repo.exists_remote(api):
            await checkout_remote(
                repo=repo, code=code, config=config, venv=venv, git=git
            )
        else:
            msg.fail(f"Project: {project!r} not found locally or on GitHub.", exits=1)
    else:
        # Unrecognised regex
        msg.fail(
            f"Argument: {project} did not match valid pattern.",
            text=(
                "Valid pattern is 'project' for a local or remote project you own,"
                " or 'user/project' to fork a repo you don't own."
            ),
            exits=1,
        )


async def checkout_fork(
    owner: str, name: str, api: API, config: Config, git: Git, venv: bool
) -> None:
    """
    Forks the passed repo, clones it, sets the upstream and informs
    the user along the way.
    """
    msg.info(f"'{owner}/{name}' belongs to {owner!r}")
    confirmed: bool = await questionary.confirm(
        f"This will fork '{owner}/{name}' to your GitHub. Are you sure?",
        default=False,
        auto_enter=False,
    ).ask_async()

    if not confirmed:
        raise click.Abort()

    # Check if we've already forked it, in which case the repo will already
    # exist under the user's namespace
    fork = Repo(
        owner=config.username,
        name=name,
        local_path=config.projects_dir.joinpath(name),
    )

    if await fork.exists_remote(api):
        msg.warn(
            f"Looks like you've already forked '{owner}/{name}'.",
            text=f"Use 'pytoil checkout {name}' to pull down your fork.",
            exits=1,
        )

    with msg.loading(f"Forking '{owner}/{name}'..."):
        try:
            await api.create_fork(owner=owner, repo=name)
        except httpx.HTTPStatusError as err:
            utils.handle_http_status_error(err)

        # Forking happens asynchronously
        # see https://docs.github.com/en/rest/reference/repos#create-a-fork
        # So here we naively just wait a few seconds before trying to clone the fork
        await asyncio.sleep(3)

    if not await fork.exists_remote(api):
        msg.warn(
            "Fork not available yet.",
            text=(
                "Forking happens asynchronously so this is normal. Give it a few"
                " more seconds and try checking it out again."
            ),
            exits=1,
        )

    msg.info(f"Cloning your fork: {config.username}/{name}.", spaced=True)
    # Only cloning 1 repo so makes sense to show the clone output
    await git.clone(url=fork.clone_url, cwd=config.projects_dir, silent=False)

    msg.info("Setting 'upstream' to original repo.")
    await git.set_upstream(owner=owner, repo=name, cwd=fork.local_path)

    # Automatic environment detection
    env = await fork.dispatch_env()

    # Special code instance because the argument 'project' will be the user/repo pattern
    code = VSCode(root=config.projects_dir.joinpath(name))

    if venv:
        await handle_venv_creation(env=env, config=config, code=code)

    if config.vscode:
        msg.info(f"Opening {fork.name!r} in VSCode.")
        await code.open()


async def handle_venv_creation(
    env: Environment | None, config: Config, code: VSCode
) -> None:
    """
    Handles automatic detection and creation of python virtual
    environments based on detected repo context.
    """
    if not env:
        msg.warn("Unable to auto-detect required environment. Skipping.")

    else:
        msg.info(f"Auto creating virtual environment using: {env.name}")
        if env.name == "conda":
            msg.text("Note: Conda environments can take a few minutes to create.")

        try:
            with msg.loading("Working..."):
                await env.install_self(silent=True)
        except ExternalToolNotInstalledError:
            msg.fail(title=f"{env.name!r} not installed", exits=1)
        except EnvironmentAlreadyExistsError:
            msg.warn(
                title="Environment already exists",
                text="No need to create a new one, Skipping.",
            )
        finally:
            # Because the first exception will exit, this is okay
            # to call as finally
            if config.vscode:
                await code.set_workspace_python(python_path=env.executable)


async def checkout_local(repo: Repo, code: VSCode, config: Config, venv: bool) -> None:
    """
    Helper to checkout a local repo.
    """
    msg.info(f"{repo.name} available locally")

    # No environment or git stuff here, chances are if it exists locally
    # user has already done all this stuff
    if venv:
        click.secho("Note: '--venv' is ignored for local projects.", fg="yellow")

    if config.vscode:
        msg.info(f"Opening {repo.name!r} in VSCode.")
        await code.open()


async def checkout_remote(
    repo: Repo, code: VSCode, config: Config, venv: bool, git: Git
) -> None:
    """
    Helper to checkout a remote repo.
    """
    msg.info(f"{repo.name} found on GitHub. Cloning...", spaced=True)
    # Only cloning 1 repo so makes sense to show output
    await git.clone(url=repo.clone_url, cwd=config.projects_dir, silent=False)

    env = await repo.dispatch_env()

    if venv:
        await handle_venv_creation(env=env, config=config, code=code)

    if config.vscode:
        msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
        await code.open()



"""
Module responsible for handling poetry environments.

Here we take advantage of poetry's new `local` config setting
to enforce the virtual environment being in the project without
altering the user's base config.


Author: Tom Fleet
Created: 24/12/2021
"""

from __future__ import annotations

import asyncio
import shutil
import sys
from pathlib import Path
from typing import Sequence

import aiofiles.os

from pytoil.exceptions import PoetryNotInstalledError

POETRY = shutil.which("poetry")


class Poetry:
    def __init__(self, root: Path, poetry: str | None = POETRY) -> None:
        self.root = root
        self.poetry = poetry

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(root={self.root!r}, poetry={self.poetry!r})"
        )

    __slots__ = ("root", "poetry")

    @property
    def project_path(self) -> Path:
        return self.root.resolve()

    @property
    def executable(self) -> Path:
        return self.project_path.joinpath(".venv/bin/python")

    @property
    def name(self) -> str:
        return "poetry"

    async def enforce_local_config(self, silent: bool = False) -> None:
        """
        Ensures any changes to poetry's config such as storing the
        virtual environment in the project directory as we do here, do not
        propegate to the user's global poetry config.
        """
        if not self.poetry:
            raise PoetryNotInstalledError

        proc = await asyncio.create_subprocess_exec(
            self.poetry,
            "config",
            "virtualenvs.in-project",
            "true",
            "--local",
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def exists(self) -> bool:
        """
        Checks whether the virtual environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then both the project and the virtual environment
        must also exist and must therefore be valid.
        """
        return await aiofiles.os.path.exists(self.executable)

    async def create(
        self, packages: Sequence[str] | None = None, silent: bool = False
    ) -> None:
        """
        This method is not implemented for poetry environments.

        Use `install` instead as with poetry, creation and installation
        are handled together.
        """
        raise NotImplementedError

    async def install(self, packages: Sequence[str], silent: bool = False) -> None:
        """
        Calls `poetry add` to install packages into the environment.

        Args:
            packages (List[str]): List of packages to install.
            silent (bool, optional): Whether to discard or display output.
        """
        if not self.poetry:
            raise PoetryNotInstalledError

        await self.enforce_local_config()

        proc = await asyncio.create_subprocess_exec(
            self.poetry,
            "add",
            *packages,
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()

    async def install_self(self, silent: bool = False) -> None:
        """
        Calls `poetry install` under the hood to install the current package
        and all it's dependencies.

        Args:
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        if not self.poetry:
            raise PoetryNotInstalledError

        await self.enforce_local_config()

        proc = await asyncio.create_subprocess_exec(
            self.poetry,
            "install",
            cwd=self.project_path,
            stdout=asyncio.subprocess.DEVNULL if silent else sys.stdout,
            stderr=asyncio.subprocess.DEVNULL if silent else sys.stderr,
        )

        await proc.wait()



from pytoil.environments.base import Environment
from pytoil.environments.conda import Conda
from pytoil.environments.flit import Flit
from pytoil.environments.poetry import Poetry
from pytoil.environments.reqs import Requirements
from pytoil.environments.virtualenv import Venv

__all__ = [
    "Poetry",
    "Environment",
    "Venv",
    "Requirements",
    "Flit",
    "Conda",
]
